// File: "jss-scmlib1.js"

// Author: Marc Feeley
//   Dec 2, 2007 : version 1.0

// Modified by: Catherine Gaudron
//  last modification: Jan 3, 2008

// miscellaneous functions

function jss_eval (str)
{ return new Function ("return " + str) ()
}

function jss_return (k,val)
{ return function () { return k (val) }
}

function jss_return_void (k)
{ return jss_return (k, false)
}

function scm_jseval (k, exp)
{ return jss_return (k, jss_eval(exp));
}

function scm_not (k,obj)
{ return jss_return (k, obj === false)
}

function scm_boolean_3f_ (k,obj)
{ return jss_return (k, typeof obj == "boolean")
}

function scm_eq_3f_ (k,obj1,obj2)
{ return jss_return (k, obj1 === obj2)
}

function scm_error (k, msg)
{ alert (msg);
  return jss_return (jss_end, null)
}

// list functions

function jss_pair (car,cdr)
{ this.car = car
  this.cdr = cdr
}

function jss_cons (car,cdr)
{ return new jss_pair (car, cdr)
}

function scm_pair_3f_ (k,obj)
{ return jss_return (k, obj instanceof jss_pair)
}

function scm_cons (k,car,cdr)
{ return jss_return (k, jss_cons (car, cdr))
}

function scm_set_2d_car_21_ (k,pair,x)
{ pair.car = x
  return jss_return_void (k)
}

function scm_set_2d_cdr_21_ (k,pair,x)
{ pair.cdr = x
  return jss_return_void (k)
}

function scm_car (k,pair) { return jss_return (k, pair.car) }
function scm_cdr (k,pair) { return jss_return (k, pair.cdr) }
function scm_caar (k,pair) { return jss_return (k, pair.car.car) }
function scm_cadr (k,pair) { return jss_return (k, pair.cdr.car) }
function scm_cdar (k,pair) { return jss_return (k, pair.car.cdr) }
function scm_cddr (k,pair) { return jss_return (k, pair.cdr.cdr) }
function scm_caaar (k,pair) { return jss_return (k, pair.car.car.car) }
function scm_caadr (k,pair) { return jss_return (k, pair.cdr.car.car) }
function scm_cadar (k,pair) { return jss_return (k, pair.car.cdr.car) }
function scm_caddr (k,pair) { return jss_return (k, pair.cdr.cdr.car) }
function scm_cdaar (k,pair) { return jss_return (k, pair.car.car.cdr) }
function scm_cdadr (k,pair) { return jss_return (k, pair.cdr.car.cdr) }
function scm_cddar (k,pair) { return jss_return (k, pair.car.cdr.cdr) }
function scm_cdddr (k,pair) { return jss_return (k, pair.cdr.cdr.cdr) }
function scm_caaaar (k,pair) { return jss_return (k, pair.car.car.car.car) }
function scm_caaadr (k,pair) { return jss_return (k, pair.cdr.car.car.car) }
function scm_caadar (k,pair) { return jss_return (k, pair.car.cdr.car.car) }
function scm_caaddr (k,pair) { return jss_return (k, pair.cdr.cdr.car.car) }
function scm_cadaar (k,pair) { return jss_return (k, pair.car.car.cdr.car) }
function scm_cadadr (k,pair) { return jss_return (k, pair.cdr.car.cdr.car) }
function scm_caddar (k,pair) { return jss_return (k, pair.car.cdr.cdr.car) }
function scm_cadddr (k,pair) { return jss_return (k, pair.cdr.cdr.cdr.car) }
function scm_cdaaar (k,pair) { return jss_return (k, pair.car.car.car.cdr) }
function scm_cdaadr (k,pair) { return jss_return (k, pair.cdr.car.car.cdr) }
function scm_cdadar (k,pair) { return jss_return (k, pair.car.cdr.car.cdr) }
function scm_cdaddr (k,pair) { return jss_return (k, pair.cdr.cdr.car.cdr) }
function scm_cddaar (k,pair) { return jss_return (k, pair.car.car.cdr.cdr) }
function scm_cddadr (k,pair) { return jss_return (k, pair.cdr.car.cdr.cdr) }
function scm_cdddar (k,pair) { return jss_return (k, pair.car.cdr.cdr.cdr) }
function scm_cddddr (k,pair) { return jss_return (k, pair.cdr.cdr.cdr.cdr) }

function scm_null_3f_ (k,obj)
{ return jss_return (k, obj === null)
}

function scm_list_3f_ (k,obj)
{ while (obj instanceof jss_pair)
    obj = obj.cdr
  return jss_return (k, obj === null)
}

function scm_list () // construct parameter list from "arguments" array
{ var n = arguments.length
  if (n < 1)
    alert ("too few parameters passed to list")
  var k = arguments[0]
  var args = null
  while (--n > 0)
    args = jss_cons (arguments[n], args)
  return jss_return (k, args)
}

function jss_length (lst)
{ var n = 0
  while (lst instanceof jss_pair)
    { lst = lst.cdr
      n++
    }
  return n
}

function scm_length (k,lst)
{ return jss_return (k, jss_length (lst))
}

function scm_append () // construct parameter list from "arguments" array
{ var n = arguments.length
  var k = arguments[0]
  //if no argument is passed, we return null (empty list)
  if (n < 2) return jss_return (k, null)
  var args = arguments[--n]
  var lst = null
  while (--n > 0)
  { lst = arguments[n]
    while (lst != null)
    { args = jss_cons(lst.car, args)
      lst = lst.cdr
    }
  }
  return jss_return (k, args)
}

function jss_reverse (lst)
{ var result = null
  while (lst instanceof jss_pair)
    { result = jss_cons (lst.car, result)
      lst = lst.cdr
    }
  return result
}

function scm_reverse (k,lst)
{ return jss_return (k, jss_reverse (lst))
}

function jss_list_tail (lst,i)
{ var result = lst
  while (i > 0)
  { lst = lst.cdr
    i--
  }
  return lst
}

function scm_list_2d_tail (k,lst,i)
{ return jss_return (k, jss_list_tail (lst,i))
}

function scm_list_2d_ref (k,lst,i)
{ return jss_return (k, jss_list_tail (lst, i).car)
}

// symbol functions

var gensym_count = 0

var jss_symbol_table = new Array ()

function jss_symbol (name)
{ this.name = name
}

function jss_make_symbol (name)
{ return new jss_symbol (name)
}

function jss_string2symbol (name)
{ var sym = jss_symbol_table[name]
  if (sym === undefined)
      { sym = jss_make_symbol (name)
        jss_symbol_table[name] = sym
      }
  return sym
}

function scm_symbol_3f_ (k,obj)
{ return jss_return (k, obj instanceof jss_symbol)
}

function scm_symbol_2d__3e_string (k,sym)
{ return jss_return (k, sym.name)
}

function scm_string_2d__3e_symbol (k,name)
{ return jss_return (k, jss_string2symbol (name))
}

function scm_gensym (k, sym)
{ var sym_start = "_";
  if (arguments.length > 1)
    sym_start += sym.name
  else
    sym_start += "j"
  return jss_return (k, jss_make_symbol (sym_start + (gensym_count++)))
}

// character functions

function jss_char (code)
{ this.code = code
}

function jss_make_char (code)
{ return new jss_char (code)
}

function scm_char_3f_ (k,obj)
{ return jss_return (k, obj instanceof jss_char)
}

function scm_integer_2d__3e_char (k,n)
{ return jss_return (k, jss_make_char (n))
}

function scm_char_2d__3e_integer (k,c)
{ return jss_return (k, c.code)
}

function scm_char_2d_alphabetic_3f_ (k, c)
{ var n = c.code
  return jss_return (k, (n >= 65 && n <= 90) || (n >= 97 && n <= 122))
}

function scm_char_2d_numeric_3f_ (k, c)
{ var n = c.code
  return jss_return (k, n >= 48 && n <= 57)
}

// string functions

function scm_string ()
{ var k = arguments[0]
  var n = arguments.length
  var result = "";
  for (i = 1; i < n; i ++)
  { result += String.fromCharCode (arguments[i].code) }
  return jss_return (k, result)
}

function scm_string_2d_length (k, str)
{ return jss_return (k, str.length)
}

function scm_substring (k, str, start, end)
{ return jss_return (k, str.substring(start, end))
}

function scm_string_3d__3f_ (k, str1, str2)
{ return jss_return (k, str1 == str2)
}

function scm_string_3f_ (k,obj)
{ return jss_return (k, typeof obj == "string")
}

function jss_append_strings (lst)
{ return jss_list2vector (lst).join ("")
}

function scm_append_2d_strings (k,lst)
{ return jss_return (k, jss_append_strings (lst))
}

function scm_string_2d_append () // construct parameter list from "arguments" array
{ var n = arguments.length
  if (n < 2)
    alert ("too few parameters passed to string-append")
  var k = arguments[0]
  var args = null
  while (--n > 0)
    args = jss_cons (arguments[n], args)
  return jss_return (k, jss_append_strings (args))
}

// number functions

function scm_exact_3f_ (k,obj)
{//////////TODO
  return jss_return (k, false)
}

function scm_number_3f_ (k,obj)
{ return jss_return (k, typeof obj == "number")
}

function jss_number2string (n,radix)
{ if (radix === undefined)
    radix = 10
  return n.toString (radix)
}

function jss_string2number (str,radix)//FIXME, parseFloat does NOT take a radix
{ if (radix === undefined)
    radix = 10
  var nb = str.split(".")
  if (nb.length == 1) return parseInt(nb[0],radix)
  else return parseFloat(str,radix)
}

function scm_number_2d__3e_string (k,n,radix)
{ return jss_return (k, jss_number2string (n, radix))
}

function scm_string_2d__3e_number (k,str,radix)
{ return jss_return (k, jss_string2number (str, radix))
}

function scm_string_2d__3e_list (k, str)
{ var n = str.length-1;
  var lst = null;
  for(i = n; i >= 0; i--)
      lst = jss_cons (jss_make_char (str.charCodeAt(i)), lst)
  return jss_return (k, lst)
}

// vector functions

function scm_vector_3f_ (k,obj)
{ return jss_return (k, obj instanceof Object && obj.constructor == Array)
}

function jss_make_vector (len,init)
{ var vect = new Array (len)
  if (init !== undefined)
    { while (len > 0)
        vect[--len] = init
    }
  return vect
}

function scm_vector () // construct parameter list from "arguments" array
{ var n = arguments.length
  if (n < 1)
    alert ("too few parameters passed to vector")
  var k = arguments[0]
  var args = null
  while (--n > 0)
    args = jss_cons (arguments[n], args)
  return jss_return (k, jss_list2vector (args))
}

function scm_make_2d_vector (k,len,init)
{ return jss_return (k, jss_make_vector (len, init))
}

function scm_vector_2d_length (k,vect)
{ return jss_return (k, vect.length)
}

function scm_vector_2d_ref (k,vect,i)
{ return jss_return (k, vect[i])
}

function scm_vector_2d_set_21_ (k,vect,i,x)
{ vect[i] = x
  return jss_return_void (k)
}

function jss_list2vector (lst)
{ var len = jss_length (lst)
  var vect = new Array (len)
  var i = 0
  while (i < len)
    { vect[i++] = lst.car
      lst = lst.cdr
    }
  return vect
}

function scm_list_2d__3e_vector (k,lst)
{ return jss_return (k, jss_list2vector (lst))
}

function jss_vector2list (vect)
{ var n = vect.length
  var lst = null
  while (--n >= 0)
    lst = jss_cons (vect[n], lst)
  return lst
}

function scm_vector_2d__3e_list (k,vect)
{ return jss_return (k, jss_vector2list (vect))
}

// control functions

function scm_procedure_3f_ (k,obj)
{ return jss_return (k, obj instanceof Function)
}

function scm_apply (k,f,lst)
{ var vect = jss_list2vector (jss_cons (k, lst))
  return function () { return f.apply (this, vect) }
}

function scm_call_2f_cc (k1,f)
{ return function ()
         { return f (k1, function (k2,r)
                         { return function ()
                                  { return k1 (r) }
                         })
         }
}
