;;; File: "jss-scmlib2.scm"

;;; Author: Marc Feeley
;;;   Dec 4, 2007 : version 1.0

;;; Scheme library functions written in Scheme.

(define (equal? x y)
  (cond ((pair? x)
         (and (pair? y)
              (equal? (car x) (car y))
              (equal? (cdr x) (cdr y))))
        (else
         (eqv? x y))))

(define (eqv? x y)
  (cond ((number? x)
         (and (number? y)
              (eq? (exact? x) (exact? y))
              (= x y)))
        ((char? x)
         (and (char? y)
              (= (char->integer x) (char->integer y))))
        (else
         (eq? x y))))

(define (member x lst)
  (if (null? lst)
      #f
      (if (equal? x (car lst))
          lst
          (member x (cdr lst)))))

(define (memv x lst)
  (if (null? lst)
      #f
      (if (eqv? x (car lst))
          lst
          (memv x (cdr lst)))))

(define (memq x lst)
  (if (null? lst)
      #f
      (if (eq? x (car lst))
          lst
          (memq x (cdr lst)))))

(define (assoc x lst)
  (if (null? lst)
      #f
      (let ((p (car lst)))
        (if (equal? x (car p))
            p
            (assoc x (cdr lst))))))

(define (assv x lst)
  (if (null? lst)
      #f
      (let ((p (car lst)))
        (if (eqv? x (car p))
            p
            (assv x (cdr lst))))))

(define (assq x lst)
  (if (null? lst)
      #f
      (let ((p (car lst)))
        (if (eq? x (car p))
            p
            (assq x (cdr lst))))))

(define (eval exp)
  (define (flatten-strings x)
    (cond ((null? x) "")
          ((pair? x) (string-append
                      (flatten-strings (car x))
                      (flatten-strings (cdr x))))
          (else x)))
  (jseval (flatten-strings
           (cons "jss_spawn (" (compile exp 'jss_end) ")"))))

(define (map f lst1 . other)
  
  (define (all pred lst)
    (if (pair? lst)
        (let ((x (car lst)))
          (and (pred x)
               (all pred (cdr lst))))
        #t))

  (define (map1 f lst)
    (if (not (pair? lst))
        '()
        (cons (f (car lst)) (map1 f (cdr lst)))))

  (define (mapn f lsts)
    (if (all pair? lsts)
        (cons (apply f (map1 car lsts))
              (mapn f (map1 cdr lsts)))
        '()))

(mapn f (cons lst1 other)))

(define (newline) ;;;;;;;TODO add port
  (write-string "\n"))

(define (write-char c) ;;;;;;;TODO add port
  (write-string (string c)))

(define (display obj) ;;;;;;;TODO add port
  (generic-write obj 'display write-string))

(define (write obj) ;;;;;;;TODO add port
  (generic-write obj 'write write-string))

(define (generic-write obj style wr-str)
  (cond ((pair? obj)
         (if (eq? style 'write)
             (wr-str "("))
         (generic-write (car obj) style wr-str)
         (let loop ((lst (cdr obj)))
           (cond ((pair? lst)
                  (if (eq? style 'write)
                      (wr-str " "))
                  (generic-write (car lst) style wr-str)
                  (loop (cdr lst)))
                 ((null? lst))
                 (else
                  (if (eq? style 'write)
                      (wr-str " . "))
                  (generic-write lst style wr-str))))
         (if (eq? style 'write)
             (wr-str ")")))
        ((vector? obj)
         (if (eq? style 'write)
             (wr-str "#"))
         (generic-write (vector->list obj) style wr-str))
        ((null? obj)
         (if (eq? style 'write)
             (wr-str "()")))
        ((boolean? obj)
         (wr-str (if obj "#t" "#f")))
        ((symbol? obj)
         (wr-str (symbol->string obj)))
        ((string? obj)
         (if (eq? style 'write)
             (wr-str "\""))
         (wr-str obj)
         (if (eq? style 'write)
             (wr-str "\"")))
        ((char? obj)
         (if (eq? style 'write)
             (wr-str "#\\"))
         (wr-str (string obj)))
        ((number? obj)
         (wr-str (number->string obj)))
        ((procedure? obj)
         (wr-str "#<procedure>"))
        (else
         (wr-str "#<unknown>"))))

(define (generic-object->string obj style)
  (let ((rev-strings '()))
    (generic-write
     obj
     style
     (lambda (str) (set! rev-strings (cons str rev-strings))))
    (append-strings (reverse rev-strings))))

(define (object->string obj)
  (generic-object->string obj 'write))

(define (display-to-string obj)
  (generic-object->string obj 'display))
