// File: "jss-threads.js"

// Author: Marc Feeley
//   Dec 4, 2007 : version 1.0

// waitq functions

function jss_wq_init (x)
{ x.jss_wq_next = x
  x.jss_wq_prev = x
  return x
}

function jss_wq_empty (x)
{ return x.jss_wq_next === x.jss_wq_prev
}

function jss_wq_deq (x)
{ var n = x.jss_wq_next
  var p = x.jss_wq_prev
  p.jss_wq_next = n
  n.jss_wq_prev = p
  return jss_wq_init (x)
}

function jss_wq_enq (q,x)
{ var p = q.jss_wq_prev
  p.jss_wq_next = x
  x.jss_wq_next = q
  q.jss_wq_prev = x
  x.jss_wq_prev = p
  return x
}

function jss_wq ()
{ jss_wq_init (this)
}

// sleepq functions

function jss_sq_init (x)
{ x.jss_sq_next = x
  x.jss_sq_prev = x
  return x
}

function jss_sq_empty (x)
{ return x.jss_sq_next === x.jss_sq_prev
}

function jss_sq_deq (x)
{ var n = x.jss_sq_next
  var p = x.jss_sq_prev
  p.jss_sq_next = n
  n.jss_sq_prev = p
  return jss_sq_init (x)
}

function jss_sq_enq (q,x)
{ var p = q.jss_sq_prev
  p.jss_sq_next = x
  x.jss_sq_next = q
  q.jss_sq_prev = x
  x.jss_sq_prev = p
  return x
}

function jss_sq ()
{ jss_sq_init (this)
}

// thread functions

function jss_thread (nextpc,name)
{ this.name = name
  this.specific = false
  this.cont = false
  this.nextpc = nextpc
  this.wakeup = 0
  this.result = false
  jss_wq_init (this)
  jss_sq_init (this)
}

function jss_make_thread (nextpc,name)
{ return new jss_thread (nextpc, name)
}

function jss_thread_name (t)
{ return t.name
}

function jss_thread_specific (t)
{ return t.specific
}

function jss_thread_specific_set (t,specific)
{ t.specific = specific
}

function jss_thread_cont (t)
{ return t.cont
}

function jss_thread_cont_set (t,cont)
{ t.cont = cont
}

function jss_thread_nextpc (t)
{ return t.nextpc
}

function jss_thread_nextpc_set (t,nextpc)
{ t.nextpc = nextpc
}

function jss_thread_wakeup (t)
{ return t.wakeup
}

function jss_thread_wakeup_set (t,wakeup)
{ t.wakeup = wakeup
}

function jss_thread_result (t)
{ return t.result
}

function jss_thread_result_set (t,result)
{ t.result = result
}

function scm_current_2d_thread (k)
{ return jss_return (k, jss_current_thread ())
}

function scm_thread_3f_ (k,obj)
{ return jss_return (k, obj instanceof jss_thread)
}

function scm_make_2d_thread (k,thunk,name)
{ return jss_return (k, jss_make_thread (function () { return thunk (jss_end) }, name))
}

function scm_thread_2d_name (k,t)
{ return jss_return (k, jss_thread_name (t))
}

function scm_thread_2d_specific (k,t)
{ return jss_return (k, jss_thread_specific (t))
}

function scm_thread_2d_specific_2d_set_21_ (k,t,specific)
{ jss_thread_specific_set (t, specific)
  return jss_return_void (k)
}

function scm_thread_2d_start_21_ (k,t)
{ jss_readyq_add (t)
  return jss_return (k, t)
}

function scm_thread_2d_yield_21_ (k)
{ //////////////////////TODO
  return jss_return_void (k)
}

function scm_thread_2d_sleep_21_ (k,timeout)
{ var wakeup = jss_timeout_to_wakeup (timeout)
  if (wakeup > jss_current_msecs ())
    { var ct = jss_current_thread ()
      jss_thread_cont_set (ct, k)
      jss_thread_nextpc_set (ct, false)
      jss_thread_wakeup_set (ct, wakeup)
      jss_wq_deq (ct)     // remove current thread from readyq
      jss_sleepq_add (ct) // add current thread to sleepq only
      return jss_schedule_next ()
    }
  return jss_return_void (k)
}

function scm_thread_2d_terminate_21_ (k,t)
{ //////////////////////TODO
  return jss_return_void (k)
}

function scm_thread_2d_join_21_ (k,t,timeout,timeoutval)
{ //////////////////////TODO
  return jss_return (k, false)
}

// mutex functions

function jss_mutex (name)
{ this.name = name
  this.specific = false
  this.locked = false
  jss_wq_init (this)
}

function jss_make_mutex (name)
{ return new jss_mutex (name)
}

function jss_mutex_name (m)
{ return m.name
}

function jss_mutex_specific (m)
{ return m.specific
}

function jss_mutex_specific_set (m,specific)
{ m.specific = specific
}

function jss_mutex_unlock (m)
{ if (m.locked)
    { var t = m.jss_wq_next
      if (t !== m)
        jss_thread_continue (t, true)
      else
        m.locked = false
    }
}

function scm_mutex_3f_ (k,obj)
{ return jss_return (k, obj instanceof jss_mutex)
}

function scm_make_2d_mutex (k,name)
{ return jss_return (k, jss_make_mutex (name))
}

function scm_mutex_2d_name (k,m)
{ return jss_return (k, jss_mutex_name (m))
}

function scm_mutex_2d_specific (k,m)
{ return jss_return (k, jss_mutex_specific (m))
}

function scm_mutex_2d_specific_2d_set_21_ (k,m,specific)
{ jss_mutex_specific_set (m, specific)
  return jss_return_void (k)
}

function scm_mutex_2d_state (k,m)
{ ////////////////TODO
  return jss_return (k, false)
}

function scm_mutex_2d_lock_21_ (k,m,timeout)
{ if (m.locked)
    { if (timeout === undefined)
        { var ct = jss_current_thread ()
          jss_thread_cont_set (ct, k)
          jss_thread_nextpc_set (ct, false)
          jss_wq_deq (ct)
          jss_wq_enq (m, ct)
          return jss_schedule_next ()
        }
      else
        { var wakeup = jss_timeout_to_wakeup (timeout)
          if (wakeup <= jss_current_msecs ())
            return jss_return (k, false)
          else
            { var ct = jss_current_thread ()
              jss_thread_cont_set (ct, k)
              jss_thread_nextpc_set (ct, false)
              jss_thread_wakeup_set (ct, wakeup)
              jss_wq_deq (ct)
              jss_wq_enq (m, ct)
              jss_sleepq_add (ct)
              return jss_schedule_next ()
            }
        }
    }
  else
    { m.locked = true
      return jss_return (k, true)
    }
}

function scm_mutex_2d_unlock_21_ (k,m,cv,timeout)
{ if (cv === undefined)
    { jss_mutex_unlock (m)
      return jss_return (k, true)
    }
  else if (timeout === undefined)
    { var ct = jss_current_thread ()
      jss_thread_cont_set (ct, k)
      jss_thread_nextpc_set (ct, false)
      jss_wq_deq (ct)
      jss_wq_enq (cv, ct)
      jss_mutex_unlock (m)
      return jss_schedule_next ()
    }
  else
    { var wakeup = jss_timeout_to_wakeup (timeout)
      if (wakeup <= jss_current_msecs ())
        return jss_return (k, false)
      else
        { var ct = jss_current_thread ()
          jss_thread_cont_set (ct, k)
          jss_thread_nextpc_set (ct, false)
          jss_thread_wakeup_set (ct, wakeup)
          jss_wq_deq (ct)
          jss_wq_enq (cv, ct)
          jss_sleepq_add (ct)
          jss_mutex_unlock (m)
          return jss_schedule_next ()
        }
    }
}

// condvar functions

function jss_condvar (name)
{ this.name = name
  jss_wq_init (this)
}

function jss_make_condvar (name)
{ return new jss_condvar (name)
}

function jss_condvar_name (cv)
{ return cv.name
}

function jss_condvar_specific (cv)
{ return cv.specific
}

function jss_condvar_specific_set (cv,specific)
{ cv.specific = specific
}

function jss_condvar_signal (cv)
{ var t = cv.jss_wq_next
  if (t !== cv)
    jss_thread_continue (t, true)
}

function jss_condvar_broadcast (cv)
{ var probe = cv.jss_wq_next
  while (probe !== cv)
    { var t = probe
      probe = probe.jss_wq_next
      jss_thread_continue (t, true)
    }
}

function scm_condition_2d_variable_3f_ (k,obj)
{ return jss_return (k, obj instanceof jss_condvar)
}

function scm_make_2d_condition_2d_variable (k,name)
{ return jss_return (k, jss_make_condvar (name))
}

function scm_condition_2d_variable_2d_name (k,m)
{ return jss_return (k, jss_condvar_name (m))
}

function scm_condition_2d_variable_2d_specific (k,m)
{ return jss_return (k, jss_condvar_specific (m))
}

function scm_condition_2d_variable_2d_specific_2d_set_21_ (k,m,specific)
{ jss_condvar_specific_set (m, specific)
  return jss_return_void (k)
}

function scm_condition_2d_variable_2d_signal_21_ (k,cv)
{ jss_condvar_signal (cv)
  return jss_return_void (k)
}

function scm_condition_2d_variable_2d_broadcast_21_ (k,cv)
{ jss_condvar_broadcast (cv)
  return jss_return_void (k)
}

// time functions

function jss_time (msecs)
{ this.msecs = msecs
}

function jss_make_time (msecs)
{ return new jss_time (msecs)
}

function jss_current_msecs ()
{ return new Date ().getTime ()
}

function jss_timeout_to_wakeup (timeout)
{ if (timeout instanceof jss_time)
    return timeout.msecs
  else
    return jss_current_msecs () + Math.round (1000 * timeout)
}

function scm_current_2d_time (k)
{ return jss_return (k, jss_make_time (jss_current_msecs ()))
}

function scm_time_3f_ (k,obj)
{ return jss_return (k, obj instanceof jss_time)
}

function scm_time_2d__3e_seconds (k,t)
{ return jss_return (k, t.msecs / 1000)
}

function scm_seconds_2d__3e_time (k,s)
{ return jss_return (k, jss_make_time (Math.round (s / 1000)))
}

// readyq and sleepq

var jss_readyq = new jss_wq ()
var jss_sleepq = new jss_sq ()

function jss_current_thread ()
{ return jss_readyq.jss_wq_next
}

function jss_readyq_add (t)
{ return jss_wq_enq (jss_readyq, t)
}

function jss_sleepq_add (t)
{ var probe = jss_sleepq.jss_sq_next
  while (probe !== jss_sleepq && probe.wakeup <= t.wakeup)
    probe = probe.jss_sq_next
  return jss_sq_enq (probe, t)
}

function jss_spawn (nextpc)
{ var t = jss_make_thread (nextpc, undefined)
  jss_readyq_add (t)
}

function jss_end (result)
{ var ct = jss_current_thread ()
  jss_thread_result_set (ct, result)
  jss_wq_deq (ct)
  return jss_schedule_next ()
}

function jss_schedule_next ()
{ var ct = jss_current_thread ()
  if (ct === jss_readyq)
    return false
  else
    return jss_thread_nextpc (ct)
}

function jss_wakeup (now)
{ var probe = jss_sleepq.jss_sq_next
  while (probe !== jss_sleepq && probe.wakeup <= now)
    { var t = probe
      probe = probe.jss_sq_next
      jss_thread_continue (t, false)
    }
}

function jss_thread_continue (t,result)
{ var cont = jss_thread_cont (t)
  jss_thread_cont_set (t, false)
  jss_thread_nextpc_set (t, jss_return (cont, result))
  jss_wq_deq (t) // the thread might be on a wait queue in
  jss_sq_deq (t) // addition to the sleepq
  jss_readyq_add (t)
}

var jss_slice = 20
var jss_min_run_delay = 10

var jss_run_timer = false

function jss_setup_delayed_run (msecs)
{ jss_cancel_delayed_run ()
  jss_run_timer = setTimeout (jss_run, msecs)
}

function jss_cancel_delayed_run ()
{ if (jss_run_timer)
    { clearTimeout (jss_run_timer)
      jss_run_timer = false
    }
}

function jss_run ()
{ // Cancel any delayed run in case jss_run was called before
  // the delay was over.

  jss_cancel_delayed_run ()

  // Wake up sleeping threads whose wakeup time has passed.

  var now = jss_current_msecs ()
  jss_wakeup (now)

  // Resume execution of first thread in the readyq, but
  // don't run too long (because some browsers put an upper limit
  // on JavaScript execution time)

  var slice_end = now + jss_slice  // run for at most a few milliseconds
  var ct = jss_current_thread ()
  if (ct !== jss_readyq)
    { // Trampoline.

      var pc = jss_thread_nextpc (ct)
      while (pc && now < slice_end)
        { // The inner loop is to avoid calling jss_current_msecs too
          // frequently.
          var n = 0
          while (pc && n++ < 200)
            pc = pc ()
          now = jss_current_msecs ()
        }

      if (pc)
        { // Save current thread's state and round-robin the
          // runnable threads.

          ct = jss_current_thread ()
          if (ct === jss_readyq)
            alert("bug")
          jss_thread_nextpc_set (ct, pc)
          jss_wq_deq (ct)
          jss_readyq_add (ct)

          // Brief pause to trick the system into thinking the
          // JavaScript code did not execute too long.

          jss_setup_delayed_run (jss_min_run_delay)

          return
        }
    }

  // If there are sleeping threads, make sure jss_run is
  // executed at the earliest wakeup time.

  var sleeper1 = jss_sleepq.jss_sq_next
  if (sleeper1 !== jss_sleepq)
    { var msecs = sleeper1.wakeup - now
      if (msecs < jss_min_run_delay)
        msecs = jss_min_run_delay
      jss_setup_delayed_run (msecs)
    }

  // If jss_setup_delayed_run was not called, then the only way to
  // resume execution is to call jss_run from within some event
  // handler (such as an onclick handler or an XMLHttpRequest
  // handler).
}
