#! /usr/bin/env gsi-script

;;;============================================================================

;;; File: "jss.scm"

;;; Author: Marc Feeley (dec 2, 2007)

;;; This is JavaScriptScheme, a Scheme to JavaScript compiler.  It
;;; supports a subset of Scheme that includes tail calls and call/cc.
;;; This file also contains a minimal web server that calls the JSS
;;; compiler to translate Scheme code embedded in HTML pages.  It replaces
;;; sections of Scheme code delimited like this
;;;
;;;   <script language="scheme">
;;;   ...Scheme code...
;;;   </script>
;;;
;;; by the equivalent JavaScript code.
;;;
;;; This program has been tested with Gambit-C 4.4.2 .
;;;
;;; Usage:
;;;
;;;   % gsi jss.scm
;;;
;;; then visit "http://localhost:8000/test.html" with a JavaScript
;;; enabled web browser.

;;;============================================================================

(include "jss-compile.scm")

;;; HTML file translation.

(define (jss-translate filename)
  (with-output-to-string
    ""
    (lambda ()
      (with-input-from-file (list path: filename eol-encoding: 'cr-lf)
        (lambda ()
          (define first? #t)
          (let loop1 ()
            (let ((s (read-line)))
              (cond ((eof-object? s))
                    ((string-ci=? s "<script language=\"scheme\">")
                     (let ((exprs
                            (let loop2 ((rev-exprs '()))
                              (let ((x (read)))
                                (if (or (eof-object? x)
                                        (eq? x '</script>))
                                    (reverse rev-exprs)
                                    (loop2 (cons x rev-exprs)))))))
                       (println "<link rel=\"stylesheet\" type=\"text/css\" href=\"jss-console.css\"></link>")
                       (for-each
                        (lambda (module)
                          (println "<script src=\"/"
                                   module
                                   ".js\"></script>"))
                        '("jss-threads"
                          "jss-io"
                          "jss-console"
                          "jss-ajax"
                          "jss-dom"
                          "jss-scmlib1"
                          "jss-init"))
                       (println
                        (javascript->html
                         (compile-scheme
                          (if first?
                              (append (get-compiler) (get-lib) exprs)
                              exprs))))
                       (set! first? #f)
                       (loop1)))
                    (else
                     (println s)
                     (loop1))))))))))

(define (javascript->html code)
  (list
   "<script language=\"javascript\">\n"
   code
   "</script>"))

(define (get-compiler)
  (call-with-input-file
      "jss-compile.scm"
    read-all))  

(define (get-lib)
  (call-with-input-file
      "jss-scmlib2.scm"
    read-all))

(define (compile-scheme exprs)
  (exec (compile `(begin ,@exprs) 'jss_end)))

(define (exec js)
  (list "jss_spawn (" js ")\n"))

;;;----------------------------------------------------------------------------

;;; Minimal web server.

(define (start-web-server)

  (define portnum 8000)

  (let ((server-port
         (open-tcp-server
          (list port-number: portnum
                eol-encoding: 'cr-lf))))

    (define (accept-connection)
      (read server-port))

    (define (read-request connection)
      (read-line connection))

    (define (handle-request request connection)

      (define (respond str)
        (display str connection)
        (close-port connection))

      (define (unhandled)
        (pp (list 'unhandled-request: request))
        (close-port connection))

      (let ((r (call-with-input-string
                   request
                 (lambda (in)
                   (read-all in (lambda (in) (read-line in #\space)))))))
        (if (pair? r)
            (cond ((string=? (car r) "GET")
                   (let* ((doc (cadr r))
                          (ext (path-extension doc)))
                     (cond ((member ext '(".html" ".htm"))
                            (let ((response
                                   (jss-translate (string-append "." doc))))
                              (respond response)))
                           ((member ext '(".js" ".css"))
                            (let ((response
                                   (call-with-input-file
                                       (string-append "." doc)
                                     (lambda (in)
                                       (read-line in #f)))))
                              (respond response)))
                           ((member ext '(".read-line"))
                            (thread-start!
                             (make-thread
                              (lambda ()
                                (let ((prompt
                                       (path-strip-extension doc)))
                                  (print prompt " ") ;; write prompt
                                  (let ((response (read-line)))
                                    (respond response)))))))
                           (else
                            (unhandled)))))
                  (else
                   (unhandled))))))

    (let loop ()
      (let* ((connection (accept-connection))
             (request (read-request connection)))
        (handle-request request connection)
        (loop)))))

(start-web-server)

;;;============================================================================
